from dataclasses import dataclass
from typing import Self

import feedparser  # type: ignore


@dataclass(init=True, repr=True, eq=True)
class Link:
    """
    Link to an article
    """

    title: str | None
    """
    The title of the link
    """
    href: str
    """
    The URL of the link
    """
    rel: str | None
    """
    The relation of the link. Can be alternate or related
    """
    mime_type: str | None
    """
    MIME type of the entity being linked, e.g., text/html
    """

    def __str__(self) -> str:
        return self.href

    @classmethod
    def from_feed_link(cls, link: feedparser.FeedParserDict) -> Self:
        """
        Create a Link object from a feedparser link entry

        :param link: The feedparser link entry
        :return: The Link object
        """
        title = link.get("title")
        href = link["href"]
        rel = link.get("rel")
        mime_type = link.get("type")
        return cls(title, href, rel, mime_type)



from enum import Enum
from typing import Self


class Category(Enum):
    """
    All the categories in the arXiv subject taxonomy, mapped to their codes.

    See [Arxiv category taxonomy](https://arxiv.org/category_taxonomy)
    """

    def __new__(cls, value: str, description: str | None = None) -> Self:
        """Allows for custom descriptions for each enum value"""
        obj = object.__new__(cls)
        obj._value_ = value
        if description is not None:
            obj.__doc__ = description
        return obj

    # Computer Science
    CS_AI = "cs.AI", "Computer Science - Artificial Intelligence"
    CS_AR = "cs.AR", "Computer Science - Hardware Architecture"
    CS_CC = "cs.CC", "Computer Science - Computational Complexity"
    CS_CE = "cs.CE", "Computer Science - Computational Engineering; Finance; and Science"
    CS_CG = "cs.CG", "Computer Science - Computational Geometry"
    CS_CL = "cs.CL", "Computer Science - Computation and Language"
    CS_CR = "cs.CR", "Computer Science - Cryptography and Security"
    CS_CV = "cs.CV", "Computer Science - Computer Vision and Pattern Recognition"
    CS_CY = "cs.CY", "Computer Science - Computers and Society"
    CS_DB = "cs.DB", "Computer Science - Databases"
    CS_DC = "cs.DC", "Computer Science - Distributed; Parallel; and Cluster Computing"
    CS_DL = "cs.DL", "Computer Science - Digital Libraries"
    CS_DM = "cs.DM", "Computer Science - Discrete Mathematics"
    CS_DS = "cs.DS", "Computer Science - Data Structures and Algorithms"
    CS_ET = "cs.ET", "Computer Science - Emerging Technologies"
    CS_FL = "cs.FL", "Computer Science - Formal Languages and Automata Theory"
    CS_GL = "cs.GL", "Computer Science - General Literature"
    CS_GR = "cs.GR", "Computer Science - Graphics"
    CS_GT = "cs.GT", "Computer Science - Game Theory"
    CS_HC = "cs.HC", "Computer Science - Human-Computer Interaction"
    CS_IR = "cs.IR", "Computer Science - Information Retrieval"
    CS_IT = "cs.IT", "Computer Science - Information Theory"
    CS_LG = "cs.LG", "Computer Science - Machine Learning"
    CS_LO = "cs.LO", "Computer Science - Logic in Computer Science"
    CS_MA = "cs.MA", "Computer Science - Multiagent Systems"
    CS_MM = "cs.MM", "Computer Science - Multimedia"
    CS_MS = "cs.MS", "Computer Science - Mathematical Software"
    CS_NA = "cs.NA", "Computer Science - Numerical Analysis"
    CS_NE = "cs.NE", "Computer Science - Neural and Evolutionary Computing"
    CS_NI = "cs.NI", "Computer Science - Networking and Internet Architecture"
    CS_OH = "cs.OH", "Computer Science - Other"
    CS_OS = "cs.OS", "Computer Science - Operating Systems"
    CS_PF = "cs.PF", "Computer Science - Performance"
    CS_PL = "cs.PL", "Computer Science - Programming Languages"
    CS_RO = "cs.RO", "Computer Science - Robotics"
    CS_SC = "cs.SC", "Computer Science - Symbolic Computation"
    CS_SD = "cs.SD", "Computer Science - Sound"
    CS_SE = "cs.SE", "Computer Science - Software Engineering"
    CS_SI = "cs.SI", "Computer Science - Social and Information Networks"
    CS_SY = "cs.SY", "Computer Science - Systems and Control"
    # Economics
    ECON_EM = "econ.EM", "Economics - Econometrics"
    ECON_GN = "econ.GN", "Economics - General Economics"
    ECON_TH = "econ.TH", "Economics - Theoretical Economics"
    # Electrical Engineering and Systems Science
    EESS_AS = "eess.AS", "Electrical Engineering and Systems Science - Audio and Speech Processing"
    EESS_IV = "eess.IV", "Electrical Engineering and Systems Science - Image and Video Processing"
    EESS_SP = "eess.SP", "Electrical Engineering and Systems Science - Signal Processing"
    EESS_SY = "eess.SY", "Electrical Engineering and Systems Science - Systems and Control"
    # Mathematics
    MATH_AC = "math.AC", "Mathematics - Commutative Algebra"
    MATH_AG = "math.AG", "Mathematics - Algebraic Geometry"
    MATH_AP = "math.AP", "Mathematics - Analysis of PDEs"
    MATH_AT = "math.AT", "Mathematics - Algebraic Topology"
    MATH_CA = "math.CA", "Mathematics - Classical Analysis and ODEs"
    MATH_CO = "math.CO", "Mathematics - Combinatorics"
    MATH_CT = "math.CT", "Mathematics - Category Theory"
    MATH_CV = "math.CV", "Mathematics - Complex Variables"
    MATH_DG = "math.DG", "Mathematics - Differential Geometry"
    MATH_DS = "math.DS", "Mathematics - Dynamical Systems"
    MATH_FA = "math.FA", "Mathematics - Functional Analysis"
    MATH_GM = "math.GM", "Mathematics - General Mathematics"
    MATH_GN = "math.GN", "Mathematics - General Topology"
    MATH_GR = "math.GR", "Mathematics - Group Theory"
    MATH_GT = "math.GT", "Mathematics - Geometric Topology"
    MATH_HO = "math.HO", "Mathematics - History and Overview"
    MATH_IT = "math.IT", "Mathematics - Information Theory"
    MATH_KT = "math.KT", "Mathematics - K-Theory and Homology"
    MATH_LO = "math.LO", "Mathematics - Logic"
    MATH_MG = "math.MG", "Mathematics - Metric Geometry"
    MATH_MP = "math.MP", "Mathematics - Mathematical Physics"
    MATH_NA = "math.NA", "Mathematics - Numerical Analysis"
    MATH_NT = "math.NT", "Mathematics - Number Theory"
    MATH_OA = "math.OA", "Mathematics - Operator Algebras"
    MATH_OC = "math.OC", "Mathematics - Optimization and Control"
    MATH_PR = "math.PR", "Mathematics - Probability"
    MATH_QA = "math.QA", "Mathematics - Quantum Algebra"
    MATH_RA = "math.RA", "Mathematics - Rings and Algebras"
    MATH_RT = "math.RT", "Mathematics - Representation Theory"
    MATH_SG = "math.SG", "Mathematics - Symplectic Geometry"
    MATH_SP = "math.SP", "Mathematics - Spectral Theory"
    MATH_ST = "math.ST", "Mathematics - Statistics Theory"
    # Physics
    # - Astrophysics
    ASTRO_PH_CO = "astro-ph.CO", "Astrophysics - Cosmology and Nongalactic Astrophysics"
    ASTRO_PH_EP = "astro-ph.EP", "Astrophysics - Earth and Planetary Astrophysics"
    ASTRO_PH_GA = "astro-ph.GA", "Astrophysics - Astrophysics of Galaxies"
    ASTRO_PH_HE = "astro-ph.HE", "Astrophysics - High Energy Astrophysical Phenomena"
    ASTRO_PH_IM = "astro-ph.IM", "Astrophysics - Instrumentation and Methods for Astrophysics"
    ASTRO_PH_SR = "astro-ph.SR", "Astrophysics - Solar and Stellar Astrophysics"
    # - Condensed Matter
    COND_MAT_DIS_NN = "cond-mat.dis-nn", "Condensed Matter - Disordered Systems and Neural Networks"
    COND_MAT_MES_HALL = "cond-mat.mes-hall", "Condensed Matter - Mesoscale and Nanoscale Physics"
    COND_MAT_MTRL_SCI = "cond-mat.mtrl-sci", "Condensed Matter - Materials Science"
    COND_MAT_OTHER = "cond-mat.other", "Condensed Matter - Other Condensed Matter"
    COND_MAT_SOFT = "cond-mat.soft", "Condensed Matter - Soft Condensed Matter"
    COND_MAT_STAT_MECH = "cond-mat.stat-mech", "Condensed Matter - Statistical Mechanics"
    COND_MAT_STR_EL = "cond-mat.str-el", "Condensed Matter - Strongly Correlated Electrons"
    COND_MAT_SUPR_CON = "cond-mat.supr-con", "Condensed Matter - Superconductivity"
    # - General Relativity
    GR_QC = "gr-qc", "General Relativity and Quantum Cosmology"
    # - High Energy Physics
    HEP_EX = "hep-ex", "High Energy Physics - Experiment"
    HEP_LAT = "hep-lat", "High Energy Physics - Lattice"
    HEP_PH = "hep-ph", "High Energy Physics - Phenomenology"
    HEP_TH = "hep-th", "High Energy Physics - Theory"
    # - Mathematical Physics
    MATH_PH = "math-ph", "Mathematical Physics"
    # - Nonlinear Sciences
    NLIN_AO = "nlin.AO", "Nonlinear Sciences - Adaptation and Self-Organizing Systems"
    NLIN_CD = "nlin.CD", "Nonlinear Sciences - Chaotic Dynamics"
    NLIN_CG = "nlin.CG", "Nonlinear Sciences - Cellular Automata and Lattice Gases"
    NLIN_PS = "nlin.PS", "Nonlinear Sciences - Pattern Formation and Solitons"
    NLIN_SI = "nlin.SI", "Nonlinear Sciences - Exactly Solvable and Integrable Systems"
    # - Nuclear Physics
    NUCL_EX = "nucl-ex", "Nuclear Experiment"
    NUCL_TH = "nucl-th", "Nuclear Theory"
    # - General Physics
    PHYSICS_ACC_PH = "physics.acc-ph", "Physics - Accelerator Physics"
    PHYSICS_AO_PH = "physics.ao-ph", "Physics - Atmospheric and Oceanic Physics"
    PHYSICS_APP_PH = "physics.app-ph", "Physics - Applied Physics"
    PHYSICS_ATM_CLUS = "physics.atm-clus", "Physics - Atomic and Molecular Clusters"
    PHYSICS_ATOM_PH = "physics.atom-ph", "Physics - Atomic Physics"
    PHYSICS_BIO_PH = "physics.bio-ph", "Physics - Biological Physics"
    PHYSICS_CHEM_PH = "physics.chem-ph", "Physics - Chemical Physics"
    PHYSICS_CLASS_PH = "physics.class-ph", "Physics - Classical Physics"
    PHYSICS_COMP_PH = "physics.comp-ph", "Physics - Computational Physics"
    PHYSICS_DATA_AN = "physics.data-an", "Physics - Data Analysis; Statistics and Probability"
    PHYSICS_ED_PH = "physics.ed-ph", "Physics - Physics Education"
    PHYSICS_FLU_DYN = "physics.flu-dyn", "Physics - Fluid Dynamics"
    PHYSICS_GEN_PH = "physics.gen-ph", "Physics - General Physics"
    PHYSICS_GEO_PH = "physics.geo-ph", "Physics - Geophysics"
    PHYSICS_HIST_PH = "physics.hist-ph", "Physics - History and Philosophy of Physics"
    PHYSICS_INS_DET = "physics.ins-det", "Physics - Instrumentation and Detectors"
    PHYSICS_MED_PH = "physics.med-ph", "Physics - Medical Physics"
    PHYSICS_OPTICS = "physics.optics", "Physics - Optics"
    PHYSICS_PLASM_PH = "physics.plasm-ph", "Physics - Plasma Physics"
    PHYSICS_POP_PH = "physics.pop-ph", "Physics - Popular Physics"
    PHYSICS_SOC_PH = "physics.soc-ph", "Physics - Physics and Society"
    PHYSICS_SPACE_PH = "physics.space-ph", "Physics - Space Physics"
    # - Quantum Physics
    QUANT_PH = "quant-ph", "Quantum Physics"
    # Quantitative Biology
    Q_BIO_BM = "q-bio.BM", "Quantitative Biology - Biomolecules"
    Q_BIO_CB = "q-bio.CB", "Quantitative Biology - Cell Behavior"
    Q_BIO_GN = "q-bio.GN", "Quantitative Biology - Genomics"
    Q_BIO_MN = "q-bio.MN", "Quantitative Biology - Molecular Networks"
    Q_BIO_NC = "q-bio.NC", "Quantitative Biology - Neurons and Cognition"
    Q_BIO_OT = "q-bio.OT", "Quantitative Biology - Other"
    Q_BIO_PE = "q-bio.PE", "Quantitative Biology - Populations and Evolution"
    Q_BIO_QM = "q-bio.QM", "Quantitative Biology - Quantitative Methods"
    Q_BIO_SC = "q-bio.SC", "Quantitative Biology - Subcellular Processes"
    Q_BIO_TO = "q-bio.TO", "Quantitative Biology - Tissues and Organs"
    # Quantitative Finance
    Q_FIN_CP = "q-fin.CP", "Quantitative Finance - Computational Finance"
    Q_FIN_EC = "q-fin.EC", "Quantitative Finance - Economics"
    Q_FIN_GN = "q-fin.GN", "Quantitative Finance - General Finance"
    Q_FIN_MF = "q-fin.MF", "Quantitative Finance - Mathematical Finance"
    Q_FIN_PM = "q-fin.PM", "Quantitative Finance - Portfolio Management"
    Q_FIN_PR = "q-fin.PR", "Quantitative Finance - Pricing of Securities"
    Q_FIN_RM = "q-fin.RM", "Quantitative Finance - Risk Management"
    Q_FIN_ST = "q-fin.ST", "Quantitative Finance - Statistical Finance"
    Q_FIN_TR = "q-fin.TR", "Quantitative Finance - Trading and Market Microstructure"
    # Statistics
    STAT_AP = "stat.AP", "Statistics - Applications"
    STAT_CO = "stat.CO", "Statistics - Computation"
    STAT_ME = "stat.ME", "Statistics - Methodology"
    STAT_ML = "stat.ML", "Statistics - Machine Learning"
    STAT_OT = "stat.OT", "Statistics - Other Statistics"
    STAT_TH = "stat.TH", "Statistics - Statistics Theory"



from .author import Author
from .category import Category
from .link import Link
from .query import Query, SortCriterion, SortBy, SortOrder, Field
from .article import Article
from .client import Client



import copy
import logging
import time
from collections.abc import Iterator
from datetime import UTC, datetime, timedelta

import feedparser  # type: ignore
import requests

from arxiv_client import Article, Query

logger = logging.getLogger(__name__)


class Client:
    """
    Structured Python3 client for the Arxiv API
    """

    base_search_url = "https://export.arxiv.org/api/query"
    """
    Base URL for Arxiv API
    """

    _session: requests.Session
    _last_request_dt: datetime | None

    def __init__(self) -> None:
        self._session = requests.Session()
        self._session.headers.update({"User-Agent": "arxiv-client-py"})
        self._last_request_dt = None

    # TODO: Consider async io
    def search(
        self,
        query: Query,
        page_size: int | None = None,
        paging_delay_ms: int = 250,
        paging_max_retries: int = 1,
    ) -> Iterator[Article]:
        """
        Search the Arxiv API.

        The paging parameters allow for the search to be broken up into smaller queries.
        This is useful for large result sets, which can be processed as a stream of smaller pages.
        In cases of failure, you can resume from last successful article processed by using
        the start parameter in the Query object.

        Raises an exception if a paging request fails after `paging_max_retries` attempts.

        :param query: The query to search with
        :param page_size: The number of results to get in each page. None will fetch all results in one chunk
        :param paging_delay_ms: The delay in milliseconds between each page request
        :param paging_max_retries: The max number of retries for each page request
        :return: The search results
        """
        logger.debug("Searching arXiv with query: %r", query)
        subquery = copy.deepcopy(query)
        if page_size is not None and page_size < query.max_results:
            subquery.max_results = page_size

        total_retrieved = 0
        while total_retrieved < (query.max_results or float("inf")):
            feed = self._get_sub_page(subquery, paging_delay_ms, paging_max_retries)
            total_retrieved += len(feed.entries)
            total_results = int(feed.feed.opensearch_totalresults)

            logger.debug("Retrieved %d of %d total articles", total_retrieved, total_results)
            for entry in feed.entries:
                yield Article.from_feed_entry(entry)

            if not feed.entries or total_retrieved >= total_results:
                return

            if page_size is not None:
                subquery.start += page_size
                if query.max_results is not None:
                    subquery.max_results = min(query.max_results - total_retrieved, page_size)

    def _get_sub_page(self, query: Query, paging_delay_ms: int, paging_max_retries: int) -> feedparser.FeedParserDict:
        """
        Get a chunk of search results from the Arxiv API.

        Will raise a RuntimeError if the chunk request fails after `chunk_max_retries` attempts.

        :param query: The query to search with
        :param paging_delay_ms: The delay in milliseconds between each chunk request
        :param paging_max_retries: The max number of retries for each chunk request
        :return: The search results
        """
        try_count = 0
        while try_count <= paging_max_retries:
            try:
                self._apply_paging_delay(paging_delay_ms)
                response = self._session.get(self.base_search_url, params=query._to_url_params())  # noqa SLF001
                self._last_request_dt = datetime.now(tz=UTC)
                response.raise_for_status()
                feed = feedparser.parse(response.content)
            except (requests.HTTPError, requests.exceptions.ConnectionError, requests.exceptions.ConnectTimeout) as e:
                logger.warning("Failed to retrieve page of articles: %s", e)
                try_count += 1
            else:
                logger.debug("Successfully retrieved page of %d articles", len(feed.entries))
                return feed

        msg = f"Failed to retrieve page of articles after {paging_max_retries} retries"
        logger.error(msg, extra={"page_query": query})
        raise RuntimeError(msg)

    def _apply_paging_delay(self, delay_ms: int) -> None:
        """
        Ensure a minimum delay of delay_ms since the last request. This is to avoid violating arXiv rate limit
        while fetching results in chunks.
        """
        if self._last_request_dt is None:
            return

        min_delay = timedelta(milliseconds=delay_ms)
        elapsed = datetime.now(tz=UTC) - self._last_request_dt
        if elapsed < min_delay:
            wait_time = (min_delay - elapsed).total_seconds()
            logger.debug("Waiting %s seconds before next request", wait_time)
            time.sleep(wait_time)



from dataclasses import dataclass
from enum import Enum
from typing import Any

from arxiv_client import Category


class SortBy(Enum):
    """
    Sort criterion for the query results

    See [the arXiv API User's Manual: sort order for return
    results](https://arxiv.org/help/api/user-manual#sort).
    """

    RELEVANCE = "relevance"
    LAST_UPDATED_DATE = "lastUpdatedDate"
    SUBMITTED_DATE = "submittedDate"


class SortOrder(Enum):
    """
    Sort order of the query results

    See [the arXiv API User's Manual: sort order for return
    results](https://arxiv.org/help/api/user-manual#sort).
    """

    ASC = "ascending"
    DESC = "descending"


@dataclass(init=False, repr=True)
class SortCriterion:
    """
    Sort criterion for the query results
    """

    sort_by: SortBy
    """
    Sort by
    """
    sort_order: SortOrder
    """
    Sort order
    """

    def __init__(self, sort_by: SortBy = SortBy.LAST_UPDATED_DATE, sort_order: SortOrder = SortOrder.DESC) -> None:
        """
        :param sort_by: The criterion to sort by
        :param sort_order: The order to sort by
        """
        self.sort_by = sort_by
        self.sort_order = sort_order


class Field(Enum):
    """
    Searchable fields and their corresponding prefix used in query string construction

    See [Arxiv Query Construction](https://info.arxiv.org/help/api/user-manual.html#query_details)
    """

    TITLE = "ti"
    AUTHOR = "au"
    ABSTRACT = "abs"
    COMMENT = "co"
    JOURNAL_REFERENCE = "jr"
    CATEGORY = "cat"
    REPORT_NUMBER = "rn"
    ID = "id"  # DEPRECATED: use `id_list` instead. Using this field will not handle article versions properly
    ALL = "all"


@dataclass(init=False, repr=True, eq=False)
class Query:
    """
    Typed wrapper for an Arxiv query. Multiples values within a single field are combined with `OR`,
    while multiple fields are combined with `AND`.

    For more advanced query logic, use the `custom_params` field to pass in a raw query string.

    See [Arxiv Query Interface](https://info.arxiv.org/help/api/user-manual.html#311-query-interface)
    [Arxiv Query Construction](https://info.arxiv.org/help/api/user-manual.html#query_details)
    """

    keywords: list[str]
    """
    Keywords to search across all fields.
    """
    title_keywords: list[str]
    """
    Title keywords to filter on.
    """
    author_names: list[str]
    """
    Author names to filter on.
    """
    categories: list[Category]
    """
    Subject categories to filter on.
    """
    abstract_keywords: list[str]
    """
    Keywords to search for in the article's abstract
    """
    comment_keywords: list[str]
    """
    Keywords to search for in the author's comment
    """
    article_ids: list[str]
    """
    ArXiv article IDs to filter on.
    """
    custom_params: str | None
    """
    Query param string for advanced users. This is helpful if query logic is not supported, e.g.,
    searching by `AND` logic between keywords.
    """
    sort_criterion: SortCriterion | None
    """
    SortBy and SortOrder for the query results. Default is to sort by LAST_UPDATED_DATE in DESC order.
    """
    start: int
    """
    Pagination parameter using zero-based indexing
    """
    max_results: int | None
    """
    The max number of results to get. Set to `None` to get maximum number of results allowed by the API.
    Requests with `max_results` set >30_000 will result in HTTP 400 error
    """

    def __init__(
        self,
        keywords: list[str] | None = None,
        title_keywords: list[str] | None = None,
        author_names: list[str] | None = None,
        categories: list[Category] | None = None,
        abstract_keywords: list[str] | None = None,
        comment_keywords: list[str] | None = None,
        article_ids: list[str] | None = None,
        custom_params: str | None = None,
        sort_criterion: SortCriterion | None = None,
        start: int = 0,
        max_results: int | None = 10,
    ) -> None:
        """
        :param keywords: Keywords to search across all fields
        :param title_keywords: Title keywords to filter on
        :param author_names: Author names to filter on
        :param categories: Categories to filter on
        :param abstract_keywords: Keywords to search for in the article's abstract
        :param comment_keywords: Keywords to search for in the author's comment
        :param article_ids: ArXiv article IDs to filter on. If you want to retrieve a specific version, simply append
            `v{version#}` to the article ID, e.g., `2103.12345v1` for version 1 of article `2103.12345`
        :param custom_params: Raw query string for advanced users
        :param sort_criterion: Sort criterion for the query results
        :param max_results: The max number of results to get
        """
        self.keywords = [] if keywords is None else keywords
        self.title_keywords = [] if title_keywords is None else title_keywords
        self.author_names = [] if author_names is None else author_names
        self.categories = [] if categories is None else categories
        self.abstract_keywords = [] if abstract_keywords is None else abstract_keywords
        self.comment_keywords = [] if comment_keywords is None else comment_keywords
        self.article_ids = [] if article_ids is None else article_ids
        self.custom_params = custom_params
        self.sort_criterion = SortCriterion() if sort_criterion is None else sort_criterion
        self.start = start
        self.max_results = max_results

    def __str__(self) -> str:
        return repr(self)

    def _to_url_params(self) -> dict[str, Any]:
        """
        Construct the full query as a Dict of query parameters, not yet URL encoded.

        See [Arxiv Query Construction](https://info.arxiv.org/help/api/user-manual.html#query_details)
        :return: The URL parameters
        """
        keywords = " OR ".join(f'{Field.ALL.value}:"{keyword}"' for keyword in self.keywords) if self.keywords else ""
        title_keywords = (
            " OR ".join(f'{Field.TITLE.value}:"{keyword}"' for keyword in self.title_keywords)
            if self.title_keywords
            else ""
        )
        authors = (
            " OR ".join([f'{Field.AUTHOR.value}:"{author}"' for author in self.author_names])
            if self.author_names
            else ""
        )
        categories = (
            " OR ".join([f"{Field.CATEGORY.value}:{category.value}" for category in self.categories])
            if self.categories
            else ""
        )
        abs_keywords = (
            " OR ".join(f'{Field.ABSTRACT.value}:"{keyword}"' for keyword in self.abstract_keywords)
            if self.abstract_keywords
            else ""
        )
        comment_keywords = (
            " OR ".join(f'{Field.COMMENT.value}:"{keyword}"' for keyword in self.comment_keywords)
            if self.comment_keywords
            else ""
        )

        partials: list[str] = []
        if keywords:
            partials.append(f"({keywords})")
        if title_keywords:
            partials.append(f"({title_keywords})")
        if authors:
            partials.append(f"({authors})")
        if categories:
            partials.append(f"({categories})")
        if abs_keywords:
            partials.append(f"({abs_keywords})")
        if comment_keywords:
            partials.append(f"({comment_keywords})")
        if self.custom_params:
            partials.append(f"({self.custom_params})")

        return {
            "search_query": " AND ".join(partials),
            "id_list": ",".join(self.article_ids),
            **(
                {"sortBy": self.sort_criterion.sort_by.value, "sortOrder": self.sort_criterion.sort_order.value}
                if self.sort_criterion is not None
                else {}
            ),
            **({"max_results": self.max_results} if self.max_results is not None else {}),
            **({"start": self.start} if self.start is not None else {}),
        }
